# Copyright 2017 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
"""Module to unit test mercury_agent.inspector.hwlib.lspci"""

import mock
import pytest

import mercury_agent.inspector.hwlib.lspci as lspci
from tests.unit.base import MercuryAgentUnitTest

EXAMPLE_LSPCI_OUTPUT = """Slot:	00:00.0
Class:	Host bridge [0600]
Vendor:	Intel Corporation [8086]
Device:	Xeon E7 v3/Xeon E5 v3/Core i7 DMI2 [2f00]
SVendor:	Intel Corporation [8086]
SDevice:	Device [0000]
Rev:	02
NUMANode:	0

Slot:	00:01.0
Class:	PCI bridge [0604]
Vendor:	Intel Corporation [8086]
Device:	Xeon E7 v3/Xeon E5 v3/Core i7 PCI Express Root Port 1 [2f02]
Rev:	02
Driver:	pcieport
Module:	shpchp
NUMANode:	0

Slot:	00:14.0
Class:	USB controller [0c03]
Vendor:	Intel Corporation [8086]
Device:	C610/X99 series chipset USB xHCI Host Controller [8d31]
SVendor:	ASUSTeK Computer Inc. [1043]
SDevice:	Device [8600]
Rev:	05
ProgIf:	30
Driver:	xhci_hcd
NUMANode:	0

Slot:	00:16.0
Class:	Communication controller [0780]
Vendor:	Intel Corporation [8086]
Device:	C610/X99 series chipset MEI Controller #1 [8d3a]
SVendor:	ASUSTeK Computer Inc. [1043]
SDevice:	Device [8600]
Rev:	05
Driver:	mei_me
Module:	mei_me
NUMANode:	0

Slot:	00:19.0
Class:	Ethernet controller [0200]
Vendor:	Intel Corporation [8086]
Device:	Ethernet Connection (2) I218-V [15a1]
SVendor:	ASUSTeK Computer Inc. [1043]
SDevice:	Device [85c4]
Rev:	05
Driver:	e1000e
Module:	e1000e
NUMANode:	0



Slot:	05:00.0
Class:	Non-Volatile memory controller [0108]
Vendor:	Samsung Electronics Co Ltd [144d]
Device:	NVMe SSD Controller SM951/PM951 [a802]
SVendor:	Samsung Electronics Co Ltd [144d]
SDevice:	Device [a801]
PhySlot:	2-1
Rev:	01
ProgIf:	02
Driver:	nvme
Module:	nvme
NUMANode:	0
"""


EXPECTED_PARSED_EXAMPLE_LSPCI_OUTPUT = [
    {
        'slot': u'00:00.0',
        'class_name': u'Host bridge',
        'class_id': u'0600',
        'vendor_name': u'Intel Corporation',
        'vendor_id': u'8086',
        'device_name': u'Xeon E7 v3/Xeon E5 v3/Core i7 DMI2',
        'device_id': u'2f00',
        'svendor_name': u'Intel Corporation',
        'svendor_id': u'8086',
        'sdevice_name': u'Device',
        'sdevice_id': u'0000',
        'revision': u'02',
    },
    {
        'slot': '00:01.0',
        'class_name': 'PCI bridge',
        'class_id': '0604',
        'vendor_name': 'Intel Corporation',
        'vendor_id': '8086',
        'device_name': 'Xeon E7 v3/Xeon E5 v3/Core i7 PCI Express Root Port 1',
        'device_id': '2f02',
        'revision': '02',
        'driver': 'pcieport',
    },
    {
        'slot': '00:14.0',
        'class_name': 'USB controller',
        'class_id': '0c03',
        'vendor_name': 'Intel Corporation',
        'vendor_id': '8086',
        'device_name': 'C610/X99 series chipset USB xHCI Host Controller',
        'device_id': '8d31',
        'svendor_name': 'ASUSTeK Computer Inc.',
        'svendor_id': '1043',
        'sdevice_name': 'Device',
        'sdevice_id': '8600',
        'revision': '05',
        'progif': '30',
        'driver': 'xhci_hcd'
    },
    {
        'slot': '00:16.0',
        'class_name': 'Communication controller',
        'class_id': '0780',
        'vendor_name': 'Intel Corporation',
        'vendor_id': '8086',
        'device_name': 'C610/X99 series chipset MEI Controller #1',
        'device_id': '8d3a',
        'svendor_name': 'ASUSTeK Computer Inc.',
        'svendor_id': '1043',
        'sdevice_name': 'Device',
        'sdevice_id': '8600',
        'revision': '05',
        'driver': 'mei_me'
    },
    {
        'slot': '00:19.0',
        'class_name': 'Ethernet controller',
        'class_id': '0200',
        'vendor_name': 'Intel Corporation',
        'vendor_id': '8086',
        'device_name': 'Ethernet Connection (2) I218-V',
        'device_id': '15a1',
        'svendor_name': 'ASUSTeK Computer Inc.',
        'svendor_id': '1043',
        'sdevice_name': 'Device',
        'sdevice_id': '85c4',
        'revision': '05',
        'driver': 'e1000e'
    },
    {
        'slot': '05:00.0',
        'class_name': 'Non-Volatile memory controller',
        'class_id': '0108',
        'vendor_name': 'Samsung Electronics Co Ltd',
        'vendor_id': '144d',
        'device_name': 'NVMe SSD Controller SM951/PM951',
        'device_id': 'a802',
        'svendor_name': 'Samsung Electronics Co Ltd',
        'svendor_id': '144d',
        'sdevice_name': 'Device',
        'sdevice_id': 'a801',
        'revision': '01',
        'driver': 'nvme',
        'progif': '02',
    },
]

FIELDS_PARSED_BY_MODULE = [
    'class_id',
    'class_name',
    'device_id',
    'device_name',
    'driver',
    'progif',
    'revision',
    'sdevice_id',
    'sdevice_name',
    'slot',
    'svendor_id',
    'svendor_name',
    'vendor_id',
    'vendor_name',
]


# Missing entries should be set to None to match unit behavior.
def _fixup_parsed_output():
    for device in EXPECTED_PARSED_EXAMPLE_LSPCI_OUTPUT:
        for field in FIELDS_PARSED_BY_MODULE:
            if field not in device.keys():
                device[field] = None


_fixup_parsed_output()


def get_fake_pcidevice_required_args(slot='00:00.0', class_id='beef',
                                     vendor_id='dead', device_id='ffff'):
    """Get a dict of args for lspci.PCIDevice"""
    return {
        'slot': slot,
        'class_id': class_id,
        'vendor_id': vendor_id,
        'device_id': device_id
    }


class MercuryMiscLspciUnitTests(MercuryAgentUnitTest):
    """Unit tests for mercury_agent.inspector.hwlib.lspci"""
    @mock.patch('mercury_agent.inspector.hwlib.lspci.subprocess.Popen')
    def setUp(self, popen_mock):
        """Setup a PCIBus object for each test."""
        popen_mock.return_value.communicate.return_value = (
            EXAMPLE_LSPCI_OUTPUT, '')
        popen_mock.return_value.returncode = 0
        self.pci_bus = lspci.PCIBus()

    def test_example_output_parsing(self):
        """Test if PCIBus/parse_lspci parsed the example output correctly."""
        assert len(self.pci_bus) == len(EXPECTED_PARSED_EXAMPLE_LSPCI_OUTPUT)

        for device in self.pci_bus:
            assert device in EXPECTED_PARSED_EXAMPLE_LSPCI_OUTPUT

    @mock.patch('mercury_agent.inspector.hwlib.lspci.subprocess.Popen')
    def test_lscpi_run_raises(self, popen_mock):
        """Test what happens when lspci returns non-zero error."""
        popen_mock.return_value.communicate.return_value = (
            EXAMPLE_LSPCI_OUTPUT, '')
        popen_mock.return_value.returncode = 1

        with pytest.raises(lspci.LSPCIError):
            lspci.lspci_run()

    def test_pci_device_raises_on_missing_arg(self):
        """Test that PCIDevice raises when missing args."""
        test_args = get_fake_pcidevice_required_args()
        # Check that test_args works.
        lspci.PCIDevice(**test_args)

        # Check key absence raises.
        for key in test_args:
            value = test_args[key]
            del test_args[key]
            with pytest.raises(lspci.LSPCIError):
                lspci.PCIDevice(**test_args)
            test_args[key] = value

    def test_pci_device__getattr__(self):
        """Test PCIDevice.__getattr__ behavior."""
        pci_device = lspci.PCIDevice(**get_fake_pcidevice_required_args())
        assert pci_device.slot == '00:00.0'
        assert pci_device.asdfjkl is None

    def test_pcibus_get_devices_by_class(self):
        """Test PCIBus.get_devices_by_class()"""
        devices = self.pci_bus.get_devices_by_class('0108')
        assert isinstance(devices, list)
        assert len(devices) == 1
        assert devices[0]['device_id'] == 'a802'

    def test_pcibus_has_device_class(self):
        """Test PCIBus.has_device_class()"""
        assert self.pci_bus.has_device_class('0108')
        assert self.pci_bus.has_device_class('0200')
        assert not self.pci_bus.has_device_class('1337')
        assert not self.pci_bus.has_device_class('beef')

    def test_pcibus_get_devices_by_vendor(self):
        """Test PCIBus.get_devices_by_vendor()"""
        intel_devices = self.pci_bus.get_devices_by_vendor('8086')
        assert isinstance(intel_devices, list)
        assert len(intel_devices) == 5
        for device in intel_devices:
            assert device['vendor_id'] == '8086'
            assert device in EXPECTED_PARSED_EXAMPLE_LSPCI_OUTPUT

        samsung_devices = self.pci_bus.get_devices_by_vendor('144d')
        assert isinstance(intel_devices, list)
        assert len(samsung_devices) == 1
        assert samsung_devices[0]['vendor_id'] == '144d'
        assert samsung_devices[0] in EXPECTED_PARSED_EXAMPLE_LSPCI_OUTPUT

        no_devices = self.pci_bus.get_devices_by_vendor('1337')
        assert isinstance(no_devices, list)
        assert len(no_devices) == 0

    def _membership_and_retrieval_test_helper(self, class_id, name):
        get_function_name = 'get_' + name + '_devices'
        get_func = getattr(self.pci_bus, get_function_name)

        has_function_name = 'has_' + name
        has_func = getattr(self.pci_bus, has_function_name)

        # Remove any of the class currently in the pci_bus object.
        to_delete = []
        for index in range(0, len(self.pci_bus)):
            if self.pci_bus[index]['class_id'] == class_id:
                to_delete.append(index)

        # Go backwards so indices remain valid after deletion.
        for index in reversed(to_delete):
            del self.pci_bus[index]

        assert not has_func()

        class_devices = get_func()
        assert isinstance(class_devices, list)
        assert len(class_devices) == 0

        self.pci_bus.append(
            lspci.PCIDevice(
                **get_fake_pcidevice_required_args(class_id=class_id)))

        assert has_func()

        class_devices = get_func()
        assert isinstance(class_devices, list)
        assert len(class_devices) == 1

    def test_pcibus_get_has_fibre_channel_devices(self):
        """Test PCIBus.get_fibre_channel_devices()"""
        """Test PCIBus.{get,has}_fibre_channel[_devices]()"""
        self._membership_and_retrieval_test_helper(lspci.FIBRE_CHANNEL,
                                                   'fibre_channel')

    def test_pcibus_get_has_ethernet_devices(self):
        """Test PCIBus.{get,has}_ethernet[_devices]()"""
        self._membership_and_retrieval_test_helper(lspci.ETHERNET_CONTROLLER,
                                                   'ethernet')

    def test_pcibus_get_has_network_devices(self):
        """Test PCIBus.{get,has}_network[_devices]()"""
        self._membership_and_retrieval_test_helper(lspci.NETWORK_CONTROLLER,
                                                   'network')

    def test_pcibus_get_has_raid_bus_devices(self):
        """Test PCIBus.{get,has}_raid_bus[_devices]()"""
        self._membership_and_retrieval_test_helper(lspci.RAID_CONTROLLER,
                                                   'raid_bus_controller')



import logging

from mercury_agent.capabilities import capability
from mercury_agent.backend_client import get_backend_client
from mercury_agent.hardware.drivers.drivers import get_subsystem_drivers
from mercury_agent.hardware.raid.abstraction.api import RAIDActions
from mercury_agent.inspector.inspect import global_device_info
from mercury_agent.inspector.inspectors.raid import raid_inspector

log = logging.getLogger(__name__)


def has_abstraction_handler():
    raid_drivers = get_subsystem_drivers('raid')
    for driver in raid_drivers:
        if isinstance(driver.handler, RAIDActions):
            return True
    return False


def update_inventory():
    backend_client = get_backend_client()
    raid_info = raid_inspector(global_device_info)
    mercury_id = global_device_info['mercury_id']

    log.debug('RAID configuration changed, updating inventory')
    backend_client.update(mercury_id, {'raid': raid_info})


def update_on_change(f):
    def wrapped_f(*args, **kwargs):
        result = f(*args, **kwargs)
        update_inventory()
        return result
    wrapped_f.__name__ = f.__name__
    wrapped_f.__doc__ = f.__doc__
    return wrapped_f


@capability('create_logical_drive',
            description='Creates a logical drive on the specified controller',
            kwarg_names=['adapter', 'level'],
            serial=True,
            dependency_callback=has_abstraction_handler,
            timeout=120
            )
@update_on_change
def abstract_create_logical_drive(adapter, level, drives=None, size=None, array=None):
    """
    :param adapter: Target adapter
    :type adapter: int
    :param level: 0, 1, 5, 6, 10, 1+0, 50, 60
    :type level: str
    :param drives: drives should be referenced as 0 based comma separated indexes, ranges,
        or a combination of both. For example::

            0, 1, 2, 3
                or
            0-3
                or
            0, 2, 4-8, 9, 10

        When using a range, both the lower and upper bounds are inclusive

        Another option is to use all or unassigned. `all` requires that all drives on the
        adapter are not part of the array. `unassigned` will select all drives not currently
        members of an array or participating as spares
    :param size: Size can be specified in bytes (int), a string using SI or IEC standards,
    a percent of total space, or a percent of free space available. If no size is listed,
    all available space will be used
    :param array: An index of an existing array we are updating
    :return:
    """
    log.info('Creating array on adapter {}: level={} drives={} size={} array={}'.format(
        adapter, level, drives, size, array
    ))

    # Right now, I assume that there is only ONE driver type, ie, one vendor type.

    raid_driver = get_subsystem_drivers('raid')[0]

    return raid_driver.handler.create_logical_drive(adapter, level, drives, size, array)


@capability('delete_logical_drive',
            description='Delete the specified logical drive',
            kwarg_names=['adapter', 'array', 'logical_drive'],
            serial=True,
            dependency_callback=has_abstraction_handler,
            timeout=120)
@update_on_change
def abstract_delete_logical_drive(adapter, array, logical_drive):
    """
    :param adapter:
    :param array:
    :param logical_drive:
    :return:
    """
    raid_driver = get_subsystem_drivers('raid')[0]
    return raid_driver.handler.delete_logical_drive(adapter, array, logical_drive)


@capability('clear_configuration',
            description='Clear all arrays from the adapter',
            kwarg_names=['adapter'],
            serial=True,
            dependency_callback=has_abstraction_handler,
            timeout=120)
@update_on_change
def abstract_clear_configuration(adapter):
    """
    :param adapter:
    :return:
    """
    raid_driver = get_subsystem_drivers('raid')[0]
    return raid_driver.handler.clear_configuration(adapter)


@capability('add_spares',
            description='Assign spare drives to the array',
            kwarg_names=['adapter', 'drives'],
            serial=True,
            dependency_callback=has_abstraction_handler,
            timeout=120)
@update_on_change
def abstract_add_spares(adapter, drives, array=None):
    """
    :param adapter:
    :param array:
    :param drives:
    :return:
    """
    raid_driver = get_subsystem_drivers('raid')[0]
    return raid_driver.handler.add_spares(adapter, drives, array)



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import time

from mercury.common.clients.rpc.backend import BackEndClient
from mercury.common.exceptions import MercuryCritical, MercuryGeneralException

from mercury_agent.capabilities import runtime_capabilities
from mercury_agent.configuration import get_configuration
from mercury_agent.pong import spawn_pong_process
from mercury_agent.register import get_dhcp_ip, register
from mercury_agent.remote_logging import MercuryLogHandler
from mercury_agent.rpc import AgentService

from mercury_agent.inspector import inspect

# Async Inspectors

from mercury_agent.inspector.inspectors.async_inspectors.lldp import \
    LLDPInspector


log = logging.getLogger(__name__)


RETRY_SECONDS = 15


class Agent(object):
    def __init__(self, configuration, logger):
        """

        :param configuration:
        """
        self.configuration = configuration
        self.agent_bind_address = configuration.agent.service_bind_address
        self.pong_bind_address = configuration.agent.pong_bind_address

        self.rpc_backend_url = self.configuration.agent.remote.backend_url
        self.log_handler = logger

        if not self.rpc_backend_url:
            raise MercuryCritical('Missing rpc backend in local configuration')

        self.backend = BackEndClient(self.rpc_backend_url,
                                     linger=0,
                                     response_timeout=10,
                                     rcv_retry=3)

    def run(self, dhcp_ip_method='simple'):
        # TODO: Add other mechanisms for enumerating the devices public ip
        log.debug('Agent: %s, Pong: %s' % (self.agent_bind_address,
                                           self.pong_bind_address))

        log.info('Running inspectors')

        device_info = inspect.inspect()

        log.info('Registering device inventory for MercuryID {}'.format(
            device_info['mercury_id']))

        log.info('Starting pong service')
        spawn_pong_process(self.pong_bind_address)

        log.info('Registering device')

        local_ip = get_dhcp_ip(device_info, method=dhcp_ip_method)

        # TODO: enumerate ipv6 addresses
        local_ipv6 = None

        while True:
            result = register(
                self.backend,
                device_info,
                local_ip,
                local_ipv6,
                runtime_capabilities)

            if result.get('error'):
                log.info('Registration was not successful, retrying...')
                time.sleep(RETRY_SECONDS)
                continue

            log.info('Device has been registered successfully')
            break

        # LogHandler

        log.info('Injecting MercuryID for remote logging')
        self.log_handler.set_mercury_id(device_info['mercury_id'])
        log.info('Injection completed')

        # AsyncInspectors
        try:
            LLDPInspector(device_info, self.backend).inspect()
        except MercuryGeneralException as mge:
            log.error(
                'Caught recoverable exception running async inspector: '
                '{}'.format(mge))

        log.info('Starting agent rpc service: %s' % self.agent_bind_address)

        agent_service = AgentService(self.agent_bind_address,
                                     self.rpc_backend_url)
        agent_service.start()


def setup_logging(configuration):
    logging.basicConfig(level=configuration.logging.level,
                        format=configuration.logging.format)

    fh = logging.FileHandler(configuration.agent.log_file)

    fh.setLevel(configuration.logging.level)
    formatter = logging.Formatter(configuration.logging.format)
    fh.setFormatter(formatter)

    mercury_logger = logging.getLogger('mercury_agent')
    mercury_logger.addHandler(fh)
    mercury_logger.info('Starting Agent')

    # Quiet these down
    logging.getLogger('mercury_agent.pong').setLevel(
        configuration.agent.pong_log_level)
    logging.getLogger('hpssa._cli').setLevel(logging.ERROR)

    # Configure the remote logging handler
    mh = MercuryLogHandler(configuration.agent.remote.log_service_url)
    mercury_logger.addHandler(mh)

    # Return this so that we can inject the mercury_id once we have it
    return mh


def main():
    configuration = get_configuration()
    mercury_handler = setup_logging(configuration)
    agent = Agent(configuration, mercury_handler)
    agent.run(configuration.agent.dhcp_ip_source)


if __name__ == '__main__':
    main()



import logging


from mercury_agent.hardware.drivers import get_subsystem_drivers
from mercury_agent.inspector.inspectors import expose_late


log = logging.getLogger(__name__)


# noinspection PyUnusedLocal
@expose_late('bmc')
def bmc_inspector(device_info):

    drivers = get_subsystem_drivers('bmc')
    if not drivers:
        return

    if len(drivers) > 1:
        log.warning('Found more than one driver for BMC. This is most likely a bug')

    driver = drivers[0]

    log.info('Running BMC inspector: {}'.format(driver.name))

    return driver.inspect()



import os
import unittest

from mercury_agent.hardware.oem.dell import om_xml_deserializer as om


def read_resource(filename):
    with open(os.path.join(os.path.dirname(__file__),
                           'resources/' + filename), 'rb') as fp:
        return fp.read()


class TestOMXML(unittest.TestCase):
    def setUp(self):
        self.chassis_xml = read_resource('chassis.xml')

    def test_loader(self):
        xl = om.XLoader(self.chassis_xml)
        self.assertEqual(xl.root.tag, 'OMA')

    def test_chassis_chassis(self):
        xl = om.XLoader(self.chassis_xml)
        chassis = om.XMLChassisStatus(xl.root)
        self.assertEqual(chassis['processor']['status_string'], 'OK')



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging

LOG = logging.getLogger(__name__)

runtime_capabilities = {}


def add_capability(entry, name, description, doc=None, serial=False,
                   num_args=None, kwarg_names=None, no_return=False,
                   dependency_callback=None, timeout=1800,
                   task_id_kwargs=False):
    """Add a new capability to the runtime capabilities.

    :param entry: The new capability.
    :param name: Name of the new capability.
    :param description: Description of the new capability.
    :param doc: Function docstring.
    :param serial: Boolean indication if the task is serial.
    :param num_args: Number of expected arguments.
    :param kwarg_names: Named arguments.
    :param no_return: True if the task doesn't return any value.
    :param dependency_callback: Callback to check dependency.
    :param timeout: Timeout for the new capability.
    :param task_id_kwargs: Whether to put task_id in kwargs.
    """
    LOG.info('Adding capability %s' % name)
    runtime_capabilities[name] = {
        'name': name,
        'entry': entry,
        'description': description,
        'doc': doc,
        'serial': serial,
        'num_args': num_args,
        'kwarg_names': kwarg_names,
        'no_return': no_return,
        'dependency_callback': dependency_callback,
        'timeout': timeout,
        'task_id_kwargs': task_id_kwargs
    }


def capability(name, description, serial=False, num_args=None,
               kwarg_names=None, no_return=False,
               dependency_callback=None, timeout=1800, task_id_kwargs=False):
    """Decorator to add a new capability.

    :param name: Name of the new capability.
    :param description: Description of the new capability.
    :param serial: Boolean indication if the task is serial.
    :param num_args: Number of expected arguments.
    :param kwarg_names: Named arguments.
    :param no_return: True if the task doesn't return any value.
    :param dependency_callback: Callback to check dependency.
    :param timeout: Timeout for the new capability.
    :param task_id_kwargs: Whether to put task_id in kwargs.
    """
    def wrap(entry):
        add_capability(entry, name, description, doc=entry.__doc__,
                       serial=serial, num_args=num_args,
                       kwarg_names=kwarg_names, no_return=no_return,
                       dependency_callback=dependency_callback,
                       timeout=timeout, task_id_kwargs=task_id_kwargs)
        return entry
    return wrap
